# -*- coding: UTF-8 -*-
#
#  Copyright (C) 2001, 2002 by Tamito KAJIYAMA
#  Copyright (C) 2002, 2003 by MATSUMURA Namihiko <nie@counterghost.net>
#  Copyright (C) 2002-2004 by Shyouzou Sugitani <shy@debian.or.jp>
#  Copyright (C) 2003 by Shun-ichi TAHARA <jado@flowernet.gr.jp>
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License (version 2) as
#  published by the Free Software Foundation.  It is distributed in the
#  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
#  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
#  PURPOSE.  See the GNU General Public License for more details.
#
# $Id: sakura.py,v 1.111 2004/04/03 06:49:25 shy Exp $
#

import os
import re
import socket
import select
import string
import sys
import time
import whrandom
import StringIO
import codecs
import types

if os.environ.has_key("DISPLAY"):
    import gtk

import ninix.main
import ninix.makoto
import ninix.script
import ninix.update
import ninix.communicate

# left mouse button
BUTTON1_CLOSE = "close"
BUTTON1_RAISE = "raise"
BUTTON1_LOWER = "lower"

# right mouse button
BUTTON3_CLOSE = "close"
BUTTON3_RAISE = "raise"
BUTTON3_LOWER = "lower"

# default browser
DEFAULT_BROWSER = "netscape -remote 'openURL(%s,new-window)'"

range_surface_scale = [("100%", 100),
                       (" 90%", 90),
                       (" 80%", 80),
                       (" 70%", 70),
                       (" 60%", 60),
                       (" 50%", 50),
                       (" 40%", 40)]

range_script_speed = [(_("None"), -1),
                      ("1 (" + _("Fast") + ")", 0),
                      ("2", 1),
                      ("3", 2),
                      ("4", 3),
                      ("5", 4),
                      ("6", 6),
                      ("7 (" + _("Slow") + ")", 8)]

class Sakura:
    BALLOON_LIFE   = 10  # sec (0: never closed automatically)
    SELECT_TIMEOUT = 15  # sec
    PAUSE_TIMEOUT  = 30  # sec
    SILENT_TIME    = 15  # sec
    # script modes
    BROWSE_MODE = 1
    SELECT_MODE = 2
    PAUSE_MODE  = 3
    # script origins
    FROM_SSTP_CLIENT = 1
    FROM_GHOST       = 2
    # HTML entity definitions
    try:
        from htmlentitydefs import name2codepoint
    except:
        name2codepoint = None
    def __init__(self, debug):
        self.debug = debug
        self.sstp_servers = []
        self.sstp_handle = None
        self.sstp_entry_db = None
        self.sstp_request_handler = None
        if self.debug & 16384:
            self.script_parser = ninix.script.Parser(error="strict")
        else:
            self.script_parser = ninix.script.Parser(error="loose")
        self.directsstp_path = None
        self.script_queue = []
        self.script_mode = self.BROWSE_MODE
        self.script_post_proc = None
        self.__balloon_life = 0
        self.__surface_life = 0
        self.surface_mouse_motion = None
        self.time_critical_session = 0
        self.user_interaction = 0
        self.lock_repaint = 0
        self.passivemode = 0
        self.running = 0
        self.event_queue = []
        self.anchor = None
        self.clock = (0, 0)
        self.surface_scale = 100 # %
        self.scale_balloon = 0
        self.balloon_scale = 100 # %
        self.script_speed = 3
        self.set_event_kill_list([])
        self.set_mouse_button1(BUTTON1_RAISE)
        self.set_mouse_button3(BUTTON3_CLOSE)
        self.set_browser(DEFAULT_BROWSER)
        self.set_helpers([])
        self.set_ghost_time(0)
        self.set_ghost_vanished_count(0)
        self.keep_silence(0)
        self.communicate = ninix.communicate.Communicate()
        self.old_otherghostname = None
        self.cantalk = 1
        self.charset = 'Shift_JIS'
        self.top_margin = 0
        self.bottom_margin = 0
    def set_ghost(self, ghost):
        self.ghost = ghost
    def set_surface(self, surface):
        self.surface = surface
    def set_balloon(self, balloon):
        self.balloon = balloon
    def set_top_margin(self, value):
        self.top_margin = value
        self.align_current()
    def set_bottom_margin(self, value):
        self.bottom_margin = value
        self.align_current()
    def add_sstp_server(self, server):
        self.sstp_servers.append(server)
    def add_directsstp_server(self, server, path):
        self.sstp_servers.append(server)
        self.directsstp_path = path
    def get_directsstp_path(self):
        return self.directsstp_path
    def finalize(self): ## FIXME
        self.save_history()
        self.communicate.rebuild_ghostdb(None)
        self.ghost.unload_shiori()
        for server in self.sstp_servers:
            server.close()
    def hide_all(self):
        self.surface.hide(0)
        self.surface.hide(1)
        self.balloon.hide(0)
        self.balloon.hide(1)
    def __raise_all(self):
        self.surface.raise_(0)
        self.surface.raise_(1)
        self.balloon.raise_(0)
        self.balloon.raise_(1)
    def __lower_all(self):
        self.surface.lower(0)
        self.surface.lower(1)
        self.balloon.lower(0)
        self.balloon.lower(1)
    def is_URL(self, s):
        return s[:7] == "http://" or s[:6] == "ftp://" or s[:6] == "file:/"
    def is_anchor(self, id):
        if len(id) == 2 and id[0] == 'anchor':
            return 1
        else:
            return 0
    def launch_browser(self, url):
        command = self.get_browser()
        self.execute_command(command, url)
    def execute_command(self, command, arg):
        pos = command.find("%s")
        if pos < 0:
            sys.stderr.write("cannot execute command (%s missing)\n")
            return
        os.system(command[:pos] + arg + command[pos+2:] + " &")
    ###   CALLBACK   ###
    def reset_idle_time(self):
        self.idle_start = time.time()
    def notify_link_selection(self, id, text, number):
        if self.script_origin == self.FROM_SSTP_CLIENT and \
           self.sstp_request_handler is not None:
            self.sstp_request_handler.send_answer(text)
            self.sstp_request_handler = None
        if self.is_anchor(id):
            self.notify_event("OnAnchorSelect", id[1])
        elif self.is_URL(id):
            self.launch_browser(id)
            self.reset_script(1)
            self.stand_by(0)
        elif self.sstp_entry_db:
            # leave the previous sstp message as it is
            self.start_script(self.sstp_entry_db.get(id, r"\e"))
            self.sstp_entry_db = None
        elif not self.notify_event("OnChoiceSelect", id, text, number):
            self.reset_script(1)
            self.stand_by(0)
    def busy(self):
        return self.time_critical_session or \
               self.user_interaction or \
               self.ghost.updateman.is_active() or \
               self.event_queue or \
               self.passivemode
    def notify_surface_drag(self, side, x_delta, y_delta):
        if self.passivemode:
            return
        x, y = self.surface.get_position(side)
        x += x_delta
        if self.surface.get_alignment(side) == 2:
            y += y_delta
        self.surface.set_position(side, x, y)
        # reset balloon positions
        self.position_balloons()
    def notify_surface_click(self, button, click, side, x, y):
        if button == 1 and self.mouse_button1 == BUTTON1_RAISE or \
           button == 3 and self.mouse_button3 == BUTTON3_RAISE:
            self.__raise_all()
        elif button == 1 and self.mouse_button1 == BUTTON1_LOWER or \
             button == 3 and self.mouse_button3 == BUTTON3_LOWER:
            self.__lower_all()
        if self.vanished:
            if side == 0 and button == 1:
                if self.sstp_request_handler:
                    self.sstp_request_handler.send_sstp_break()
                    self.sstp_request_handler = None
                self.reset_script(1)
                self.notify_event("OnVanishButtonHold", default=r"\e")
                self.vanished = 0
            return
        if self.ghost.updateman.is_active():
            if button == 1 and click == 2:
                self.ghost.updateman.interrupt()
            return
        if self.time_critical_session:
            return
        elif button == 1 and click == 1:
            if self.passivemode and self.processed_script != None:
                return
            part = self.surface.get_touched_region(side, x, y)
            self.notify_event("OnMouseClick", x, y, "", side, part)
        elif self.passivemode:
            return
        elif button == 3:
            if self.sstp_request_handler:
                self.sstp_request_handler.send_sstp_break()
                self.sstp_request_handler = None
                self.reset_script(1)
                self.stand_by(0)
            self.surface.open_popup_menu(button, side)
        elif button == 1 and click == 2:
            if self.sstp_request_handler:
                self.sstp_request_handler.send_sstp_break()
                self.sstp_request_handler = None
            part = self.surface.get_touched_region(side, x, y)
            self.notify_event("OnMouseDoubleClick", x, y, "", side, part)
    def notify_surface_mouse_motion(self, side, x, y):
        if self.surface_mouse_motion is not None:
            return
        part = self.surface.get_touched_region(side, x, y)
        if part:
            self.surface_mouse_motion = (side, x, y, part)
        else:
            self.surface_mouse_motion = None
    def notify_surface_mouse_scroll(self, side, x, y, direction):
        if direction == 'up':
            part = self.surface.get_touched_region(side, x, y)
            self.notify_event("OnMouseWheel", x, y, 1, side, part)
        elif direction == 'down':
            part = self.surface.get_touched_region(side, x, y)
            self.notify_event("OnMouseWheel", x, y, -1, side, part)
    def notify_balloon_click(self, button, click, side):
        if self.script_mode == self.PAUSE_MODE:
            self.script_mode = self.BROWSE_MODE
            self.balloon.window[self.script_side].redraw() # erase down arrow
        elif button == 1 and self.mouse_button1 == BUTTON1_RAISE or \
             button == 3 and self.mouse_button3 == BUTTON3_RAISE:
            self.__raise_all()
        elif button == 1 and self.mouse_button1 == BUTTON1_LOWER or \
             button == 3 and self.mouse_button3 == BUTTON3_LOWER:
            self.__lower_all()
        if self.vanished:
            return
        if self.ghost.updateman.is_active():
            if button == 1 and click == 2:
                self.ghost.updateman.interrupt()
            return
        if self.time_critical_session:
            self.time_critical_session = not self.time_critical_session
            return
        elif self.passivemode:
            return
        elif button == 1 and click == 2:
            if self.sstp_request_handler:
                self.sstp_request_handler.send_sstp_break()
                self.sstp_request_handler = None
            self.notify_event("OnMouseDoubleClick", "", "", "", side, "")
        elif button == 1 and self.mouse_button1 == BUTTON1_CLOSE or \
             button == 3 and self.mouse_button3 == BUTTON3_CLOSE:
            if self.sstp_request_handler:
                self.sstp_request_handler.send_sstp_break()
                self.sstp_request_handler = None
            self.reset_script(1)
            self.stand_by(0)
    def notify_file_drop(self, list, side):
        event = None
        for path in list:
            if os.path.isdir(path) and event is None:
                event = "OnDirectoryDrop"
            elif os.path.isfile(path):
                for regex, command in self.helpers:
                    if regex.search(path):
                        event = "OnFileDropping"
                        self.execute_command(command, path)
                        break
        if event is None:
            event = "OnFileDropped"
        self.enqueue_event(event, path, side)
    def notify_user_communicate(self, data):
        if data is not None:
            self.notify_event("OnCommunicate", "user", data)
    def notify_user_teach(self, data):
        if data is not None:
            script = self.ghost.teach(data)
            if script:
                self.start_script(script)
                self.balloon.hide_sstp_message()
        self.user_interaction = 0
    def notify_user_input(self, symbol, data):
        if data is None:
            data = ''
        ## CHECK: symbol
        if symbol == "OnUserInput" and self.notify_event("OnUserInput", data):
            pass
        elif self.notify_event("OnUserInput", symbol, data):
            pass
        elif self.notify_event(symbol, data):
            pass
        self.user_interaction = 0
    month_names = ["Jan", "Feb", "Mar", "Apr", "May", "Jun",
                   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]
    def notify_event(self, event,
                     ref0=None, ref1=None, ref2=None, ref3=None,
                     ref4=None, ref5=None, ref6=None, ref7=None,
                     ref_ext=[], type='GET',
                     default=None):
        #if self.time_critical_session:
        #    return 0
        if event in self.event_kill_list:
            return 0
        script = self.ghost.get_event_response(
            event, ref0, ref1, ref2, ref3, ref4, ref5, ref6, ref7, ref_ext, type) or default
        if self.debug & 1 and script or \
           self.debug & 2 and not script and event != "OnSecondChange":
            t = time.localtime(time.time())
            m = self.month_names[t[1]-1]
            print "\n[%02d/%s/%d:%02d:%02d:%02d %+05d]" % (
                t[2], m, t[0], t[3], t[4], t[5], -time.timezone/36)
            print "Event:", event
            n = 0
            for value in [ref0, ref1, ref2, ref3, ref4, ref5, ref6, ref7]:
                if value is not None:
                    print "Reference%d:" % n, '%s' % str(value).encode('utf-8', 'ignore')
                n += 1
        if not script: # an empty script is ignored
            return 0
        if self.debug & 1:
            print '=> "%s"' % script.encode('utf-8', 'ignore')
        if self.passivemode and \
           (event == "OnSecondChange" or event == "OnMinuteChange"):
            return 0
        self.start_script(script)
        self.balloon.hide_sstp_message()
        return 1
    def get_event_response(self, event,
                           ref0=None, ref1=None, ref2=None, ref3=None,
                           ref4=None, ref5=None, ref6=None, ref7=None,
                           type='GET'):
        return self.ghost.get_event_response(
            event, ref0, ref1, ref2, ref3, ref4, ref5, ref6, ref7, type)
    def enqueue_script(self, script, sender, handle,
                       host, show_sstp_marker, use_translator,
                       db=None, request_handler=None):
        if not self.script_queue and not self.time_critical_session and not self.passivemode:
            if self.sstp_request_handler:
                self.sstp_request_handler.send_sstp_break()
                self.sstp_request_handler = None
            self.reset_script(1)
        self.script_queue.append((script, sender, handle, host,
                                  show_sstp_marker, use_translator,
                                  db, request_handler))
    def enqueue_event(self, event,
                      ref0=None, ref1=None, ref2=None, ref3=None,
                      ref4=None, ref5=None, ref6=None, ref7=None,
                      proc=None):
        self.event_queue.append(
            (event, ref0, ref1, ref2, ref3, ref4, ref5, ref6, ref7, proc))
    EVENT_SCRIPTS = {
        "OnUpdateBegin":    r"\t\h\s[0]" + unicode(_("Network Update has begun."), 'utf-8') + r"\e",
        "OnUpdateComplete": r"\t\h\s[5]" + unicode(_("Network Update completed successfully."), 'utf-8') + r"\e",
        "OnUpdateFailure":  r"\t\h\s[4]" + unicode(_("Network Update failed."), 'utf-8') + r"\e",
        }
    def handle_event(self):
        while self.event_queue:
            event, ref0, ref1, ref2, ref3, ref4, ref5, ref6, ref7, proc = \
                self.event_queue.pop(0)
            if self.notify_event(event,
                                 ref0, ref1, ref2, ref3,
                                 ref4, ref5, ref6, ref7,
                                 default=self.EVENT_SCRIPTS.get(event)):
                self.script_post_proc = proc
                return 1
            elif proc is not None:
                proc()
                return 1
        return 0
    def get_default_surface_scale(self):
        return range_surface_scale[0]
    def set_scale_balloon(self, scale):
        if scale:
            self.scale_balloon = 1
            self.balloon_scale = self.surface_scale
        else:
            self.scale_balloon = 0
            self.balloon_scale = 100 # %
        if self.running:
            self.position_all() ## FIXME: reset balloons
    def set_surface_scale(self, scale):
        if self.surface_scale == scale:
            return
        self.surface_scale = scale # %
        if self.scale_balloon:
            self.balloon_scale = scale # %
        if self.running:
            self.surface.reset_surface(0)
            self.surface.reset_surface(1)
            self.position_all()
    def get_surface_scale(self):
        return self.surface_scale
    def get_default_script_speed(self):
        number = len(range_script_speed)/2
        return range_script_speed[number]
    def set_script_speed(self, speed):
        if self.script_speed == speed:
            return
        self.script_speed = speed  # ordinal (-1: no wait)
    def get_script_speed(self):
        return self.script_speed
    def set_event_kill_list(self, list):
        ##print "Sakura.set_event_kill_list:", list
        self.event_kill_list = list
    def get_event_kill_list(self):
        return self.event_kill_list
    def set_mouse_button1(self, name):
        ##print "Sakura.set_mouse_button1:", name
        self.mouse_button1 = name
    def get_mouse_button1(self):
        return self.mouse_button1
    def set_mouse_button3(self, name):
        ##print "Sakura.set_mouse_button3:", name
        self.mouse_button3 = name
    def get_mouse_button3(self):
        return self.mouse_button3
    def set_browser(self, command):
        self.browser = command
    def get_browser(self):
        return self.browser
    def set_helpers(self, list):
        ##print "Sakura.set_helpers:", list
        self.helpers = []
        for pattern, command in list:
            try:
                regex = re.compile(pattern)
            except re.error:
                continue
            self.helpers.append((regex, command))
    def get_helpers(self):
        buffer = []
        for regex, command in self.helpers:
            buffer.append((regex.pattern, command))
        return buffer
    def get_ghost_time(self):
        return self.ghost_time
    def set_ghost_time(self, clock):
        self.ghost_time = clock
    def get_ghost_vanished_count(self):
        return self.ghost_vanished_count
    def set_ghost_vanished_count(self, count):
        self.ghost_vanished_count = count
    def get_username(self):
        return self.ghost.getstring("username") or self.surface.get_username() or self.ghost.get_username()
    def get_selfname(self):
        return self.surface.get_selfname() or self.ghost.get_selfname()
    def get_selfname2(self):
        return self.surface.get_selfname2() or self.ghost.get_selfname2()
    def get_keroname(self):
        return self.surface.get_keroname() or self.ghost.get_keroname()
    def get_friendname(self):
        return self.surface.get_friendname() or self.ghost.get_friendname()
    def get_sstp_port(self):
        if not self.sstp_servers:
            return None
        return self.sstp_servers[0].server_address[1]
    def keep_silence(self, quiet):
        if quiet:
            self.silent_time = time.time()
        else:
            self.silent_time = 0
            self.reset_idle_time()
    def reload(self):
        self.ghost.reload()
    def get_uptime(self):
        uptime = int(time.time() - self.start_time) / 3600
        if uptime < 0:
            self.start_time = time.time()
            return 0
        return uptime
    ###   STARTER   ###
    def stand_by(self, reset_surface):
        self.balloon.hide(0)
        self.balloon.hide(1)
        self.balloon.hide_sstp_message()
        if reset_surface:
            self.surface.set_surface(0, "0")
            self.surface.set_surface(1, "10")
            self.notify_event("OnSurfaceChange",
                              self.surface.get_surface(0),
                              self.surface.get_surface(1))
            self.balloon.set_balloon(0, 0)
            self.balloon.set_balloon(1, 0)
        elif self.surface.get_surface(0) != "0" or \
             self.surface.get_surface(1) != "10":
            self.__surface_life = whrandom.randint(20, 30)
            ##print "surface_life =", self.__surface_life
    def align_top(self, side):
        sx, sy = self.surface.get_position(side)
        sy = 0 + self.top_margin
        self.surface.set_position(side, sx, sy)
        self.surface.set_alignment(side, 1)
        self.position_balloons()
    def align_bottom(self, side):
        scrn_h = gtk.gdk.screen_height() - self.bottom_margin
        sw, sh = self.surface.get_surface_size(side)
        sx, sy = self.surface.get_position(side)
        sy = scrn_h - sh
        self.surface.set_position(side, sx, sy)
        self.surface.set_alignment(side, 0)
        self.position_balloons()
    def align_current(self):
        if len(self.surface.window) <= 0:
            return
        if self.surface.get_alignment(0) == 1: # top
            self.align_top(0)
        else:
            self.align_bottom(0)
        if self.surface.get_alignment(1) == 1: # top
            self.align_top(1)
        else:
            self.align_bottom(1)
    def position_all(self):
        scrn_w = gtk.gdk.screen_width()
        scrn_h = gtk.gdk.screen_height() - self.bottom_margin
        s0w, s0h = self.surface.get_surface_size(0)
        s1w, s1h = self.surface.get_surface_size(1)
        b0w, b0h = self.balloon.get_balloon_size(0)
        b1w, b1h = self.balloon.get_balloon_size(1)
        o0x, o0y = self.surface.get_balloon_offset(0)
        o1x, o1y = self.surface.get_balloon_offset(1)
        # sakura
        s0x = scrn_w - s0w
        if self.surface.get_alignment(0) == 1: # top
            s0y = 0 + self.top_margin
        else:
            s0y = scrn_h - s0h
        # unyuu
        offset = max(0, b1w - (b0w - o0x))
        if (s0x + o0x - b0w) - offset - s1w + o1x < 0:
            s1x = 0
        else:
            s1x = (s0x + o0x - b0w) - offset - s1w + o1x
        if self.surface.get_alignment(1) == 1: # top
            s1y = 0 + self.top_margin
        else:
            s1y = scrn_h - s1h
        # position all windows
        self.surface.set_position(0, s0x, s0y)
        self.surface.set_position(1, s1x, s1y)
        self.position_balloons()
    def position_balloons(self):
        scrn_w = gtk.gdk.screen_width()
        scrn_h = gtk.gdk.screen_height() - self.bottom_margin
        s0x, s0y = self.surface.get_position(0)
        o0x, o0y = self.surface.get_surface_offset(0)
        s0x += o0x
        s0y += o0y
        s1x, s1y = self.surface.get_position(1)
        o1x, o1y = self.surface.get_surface_offset(1)
        s1x += o1x
        s1y += o1y
        s0w, s0h = self.surface.get_surface_size(0)
        s1w, s1h = self.surface.get_surface_size(1)
        b0w, b0h = self.balloon.get_balloon_size(0)
        b1w, b1h = self.balloon.get_balloon_size(1)
        o0x, o0y = self.surface.get_balloon_offset(0)
        o1x, o1y = self.surface.get_balloon_offset(1)
        # sakura's balloon
        if s0x + s0w / 2 > s1x + s1w / 2:
            l_space = s0x - (s1x + s1w)
            r_space = scrn_w - (s0x + s0w)
        else:
            l_space = s0x
            r_space = s1x - (s0x + s0w)
        if s0x + s0w / 2 < s1x + s1w / 2:
            b0side = 1
        else:
            b0side = 0
        if b0side == 1 and b0w - o0x > r_space and b0w + o0x <= l_space:
            b0side = 0
        elif b0side == 0 and b0w + o0x > l_space and b0w - o0x <= r_space:
            b0side = 1
        if b0side == 0: # left
            b0x = max(s0x - b0w + o0x, 0)
        else:
            b0x = min(s0x + s0w - o0x, scrn_w - b0w)
        align = self.surface.get_alignment(0)
        if align == 0:
            b0y = min(s0y + o0y, scrn_h - b0h)
        elif align == 1:
            b0y = max(s0y + o0y, 0 + self.top_margin)
        else:
            if s0y < scrn_h/2:
                b0y = max(s0y + o0y, 0 + self.top_margin)
            else:
                b0y = min(s0y + o0y, scrn_h - b0h)
        # unyuu's balloon
        if s0x + s0w / 2 > s1x + s1w / 2:
            l_space = s1x
            r_space = s0x - (s1x + s1w)
        else:
            l_space = s1x - (s0x + s0w)
            r_space = scrn_w - (s1x + s1w)
        if s0x + s0w / 2 > s1x + s1w / 2:
            b1side = 1
        else:
            b1side = 0
        if b1side == 1 and b1w - o1x > r_space and b1w + o1x <= l_space:
            b1side = 0
        elif b1side == 0 and b1w + o1x > l_space and b1w - o1x <= r_space:
            b1side = 1
        if b1side == 0: # left
            b1x = max(s1x - b1w + o1x, 0)
        else:
            b1x = min(s1x + s1w - o1x, scrn_w - b1w)
        align = self.surface.get_alignment(1)
        if align == 0:
            b1y = min(s1y + o1y, scrn_h - b0h)
        elif align == 1:
            b1y = max(s1y + o1y, 0 + self.top_margin)
        else:
            if s0y < scrn_h/2:
                b1y = max(s1y + o1y, 0 + self.top_margin)
            else:
                b1y = min(s1y + o1y, scrn_h - b0h)
        # move balloons if overlapped
        if b0x <= b1x and b0x + b0w >= b1x or \
            b1x <= b0x and b1x + b1w >= b0x:
            if b0y <= b1y:
                overlap = (b0y + b0h) - b1y
                if overlap > 0:
                    space = scrn_h - (b1y + b1h)
                    if space > 0:
                        b1y = b1y + min(space, overlap)
                overlap = (b0y + b0h) - b1y
                if overlap > 0:
                    b0y = max(b0y - overlap, 0 + self.top_margin)
            else:
                overlap = (b1y + b1h) - b0y
                if overlap > 0:
                    space = scrn_h - (b0y + b0h)
                    if space > 0:
                        b0y = b0y + min(space, overlap)
                overlap = (b1y + b1h) - b0y
                if overlap > 0:
                    b1y = max(b1y - overlap, 0 + self.top_margin)
        # position balloon windows
        self.balloon.set_position(0, b0x, b0y)
        self.balloon.set_position(1, b1x, b1y)
        self.surface.set_direction(0, b0side)
        self.surface.set_direction(1, b1side)
        self.balloon.set_direction(0, b0side)
        self.balloon.set_direction(1, b1side)
    def start(self):
        self.start_time = time.time()
        self.restart()
        if self.ghost_time == 0:
            if not self.notify_event("OnFirstBoot", self.ghost_vanished_count):
                self.notify_event("OnBoot", default=ninix.main.VERSION_INFO)
        else:
            self.notify_event("OnBoot", default=ninix.main.VERSION_INFO)
        self.enqueue_event("OnDisplayChange", gtk.gdk.visual_get_best_depth(), gtk.gdk.screen_width(), gtk.gdk.screen_height())
    def save_history(self):
        path = os.path.join(self.ghost.get_prefix(), 'HISTORY')
        try:
            file = open(path, "w")
        except IOError, (code, message):
            sys.stderr.write("cannot write %s\n" % path)
        else:
            time = self.get_ghost_time()
            vanished_count = self.get_ghost_vanished_count()
            file.write('time, %s\n' % time)
            file.write('vanished_count, %s\n' % vanished_count)
            file.close()
    def restart(self):
        path = os.path.join(self.ghost.get_prefix(), 'HISTORY')
        if os.path.exists(path):
            ghost_time = 0
            ghost_vanished_count = 0
            try:
                file = open(path, "r")
            except IOError, (code, message):
                sys.stderr.write("cannot read %s\n" % path)
            else:
                for line in file:
                    comma = line.find(',')
                    if comma >= 0:
                        key = line[:comma].strip()
                        value = line[comma+1:].strip()
                    if key == 'time':
                        try:
                            ghost_time = int(value)
                        except:
                            pass
                    elif key == 'vanished_count':
                        try:
                            ghost_vanished_count = int(value)
                        except:
                            pass
                file.close()
            self.set_ghost_time(ghost_time)
            self.set_ghost_vanished_count(ghost_vanished_count)
        else:
            self.set_ghost_time(0)
            self.set_ghost_vanished_count(0)
        self.vanished = 0
        self.reset_script(1)
        self.surface.reset_alignment()
        self.stand_by(1)
        self.position_all()
        self.reset_idle_time()
        self.running = 1
        self.charset = 'Shift_JIS' # default
        charset = self.get_event_response("charset")
        if charset:
            try:
                codecs.lookup(charset)
            except:
                sys.stderr.write("Unsupported charset %s" % repr(charset))
            else:
                self.charset = charset
        else:
            pass
    def stop(self):
        self.running = 0
        self.save_history()
        self.communicate.rebuild_ghostdb(None)
        self.hide_all()
    def notify_close(self, proc):
        self.reset_script(1)
        self.enqueue_event("OnClose", proc=proc)
    def notify_ghost_changing(self, name, method, proc):
        self.reset_script(1)
        self.enqueue_event("OnGhostChanging", name, method, proc=proc)
    def notify_shell_changing(self, name, proc):
        self.reset_script(1)
        self.enqueue_event("OnShellChanging", name, proc=proc)
    def notify_vanish_selecting(self):
        self.reset_script(1)
        self.notify_event("OnVanishSelecting")
    def notify_vanish_selected(self, proc):
        self.reset_script(1)
        self.enqueue_event("OnVanishSelected", proc=proc)
        self.vanished = 1
    def notify_vanish_cancel(self):
        self.reset_script(1)
        self.notify_event("OnVanishCancel")
    def notify_ghost_changed(self, name, vanished=0):
        if self.ghost_time == 0:
            if self.notify_event("OnFirstBoot", self.ghost_vanished_count):
                return 
        elif vanished:
            if self.notify_event("OnVanished", name):
                return
        else:
            if self.notify_event("OnGhostChanged", name):
                return
        self.notify_event("OnBoot", default=ninix.main.VERSION_INFO)
    def notify_shell_changed(self, name):
        self.notify_event("OnShellChanged", name)
    def notify_ninix_reloading(self, proc):
        self.reset_script(1)
        self.enqueue_event("OnNinixReloading", proc=proc)
    def notify_ninix_reloaded(self):
        self.notify_event("OnNinixReloaded")
        self.notify_event("OnBoot", default=ninix.main.VERSION_INFO)
    def notify_about(self):
        self.start_script(ninix.main.VERSION_INFO)

    def update_surface(self):
        if self.running:
            self.surface.update()
    def receive_sstp_request(self): # enqueue scripts
        try:
            for sstp_server in self.sstp_servers:
                sstp_server.handle_request()
        except socket.error, (code, message):
            sys.stderr.write("socket.error: %s (%d)\n" % (message, code))
        except ValueError: # may happen when ninix is terminated
            return
    def process_script(self):
        now = time.time()
        idle = now - self.idle_start
        minute, second = time.localtime(now)[4:6]
        if self.clock[0] != second:
            self.ghost_time += 1
            current_ghost = (self.get_selfname(),
                             self.surface.get_surface(0),
                             self.surface.get_surface(1))
            self.communicate.rebuild_ghostdb(current_ghost)
            otherghostname = self.communicate.get_otherghostname()
            if otherghostname != self.old_otherghostname:
                args = []
                args.extend(otherghostname[:8])
                if len(otherghostname) > 8:
                    ext = otherghostname[8:]
                    args = otherghostname[:8]
                    args.append(ext)
                args.insert(0, 'otherghostname')
                args = tuple(args)
                keyword = {'type': 'NOTIFY'}
                apply(self.notify_event, args, keyword)
            self.old_otherghostname = otherghostname
        mikire, kasanari = self.surface.get_mikire_kasanari()
        if not self.running:
            pass
        elif self.script_mode == self.PAUSE_MODE:
            ##if idle > self.PAUSE_TIMEOUT:
            ##    self.script_mode = self.BROWSE_MODE
            pass
        elif self.processed_script or self.processed_text:
            self.interpret_script()
        elif self.script_post_proc is not None:
            self.script_post_proc()
            self.script_post_proc = None
        elif self.script_mode == self.SELECT_MODE:
            if self.passivemode:
                pass
            elif idle > self.SELECT_TIMEOUT:
                self.script_mode = self.BROWSE_MODE
                if self.sstp_request_handler:
                    self.sstp_request_handler.send_timeout()
                    self.sstp_request_handler = None
                if not self.notify_event("OnChoiceTimeout"):
                    self.stand_by(0)
        elif self.sstp_handle is not None:
            self.close_sstp_handle()
        elif self.user_interaction:
            pass
        elif idle > self.__balloon_life > 0 and not self.passivemode:
            self.__balloon_life = 0
            self.stand_by(0)
        elif self.event_queue and self.handle_event():
            pass
        elif self.script_queue and not self.passivemode:
            if self.silent_time > 0:
                self.keep_silence(1) # extend silent time
            script, sender, self.sstp_handle, \
                    host, show_sstp_marker, use_translator, \
                    self.sstp_entry_db, self.sstp_request_handler = \
                    self.script_queue.pop(0)
            if show_sstp_marker:
                self.balloon.show_sstp_message(sender, host)
            else:
                self.balloon.hide_sstp_message()
            # XXX: how about the use_translator flag?
            self.start_script(script, self.FROM_SSTP_CLIENT)
        elif self.silent_time > 0:
            if now - self.silent_time > self.SILENT_TIME:
                self.keep_silence(0)
        elif self.clock[0] != second and \
             self.notify_event("OnSecondChange", self.get_uptime(),
                               mikire, kasanari, not self.passivemode and self.cantalk):
            pass
        elif self.clock[1] != minute and \
             self.notify_event("OnMinuteChange", self.get_uptime(),
                               mikire, kasanari, not self.passivemode and self.cantalk):
            pass
        elif self.surface_mouse_motion is not None:
            side, x, y, part = self.surface_mouse_motion
            self.notify_event("OnMouseMove", x, y, "", side, part)
            self.surface_mouse_motion = None
        elif idle > self.__surface_life > 0 and not self.passivemode:
            self.__surface_life = 0
            self.notify_event("OnSurfaceRestore",
                              self.surface.get_surface(0),
                              self.surface.get_surface(1))
        self.clock = (second, minute)
    ###   SCRIPT PLAYER   ###
    def start_script(self, script, origin=None):
        if not script:
            return
        self.script_origin = origin or self.FROM_GHOST
        self.reset_script(1)
        if script.rstrip()[-2:] != r"\e":
            script += r"\e"
        try:
            self.processed_script = self.script_parser.parse(script)
        except ninix.script.ParserError, e:
            self.processed_script = None
            if self.sstp_request_handler:
                self.sstp_request_handler.send_no_content()
                self.sstp_request_handler = None
            print "-" * 50
            print e
            print script.encode('utf-8', 'ignore')
        self.script_mode = self.BROWSE_MODE
        self.script_wait = None
        self.script_side = 0
        self.time_critical_session = 0
        self.quick_session = 0
        self.set_synchronized_session(0)
        self.text_count = [0, 0]
        self.balloon.hide(0)
        self.balloon.hide(1)
        self.balloon.clear_text(0)
        self.balloon.clear_text(1)
        self.balloon.set_balloon(0, 0)
        self.balloon.set_balloon(1, 0)
        self.current_time = time.localtime(time.time())
        self.reset_idle_time()
    def interpret_script(self):
        if self.script_wait is not None:
            if time.time() < self.script_wait:
                return
            self.script_wait = None
        if self.processed_text:
            self.balloon.show(self.script_side)
            self.balloon.append_text(self.script_side, self.processed_text[0])
            self.processed_text = self.processed_text[1:]
            self.text_count[self.script_side] = self.text_count[self.script_side] + 1
            win = self.surface.window[self.script_side]
            id = self.surface.get_surface(self.script_side)
            count = None
            for actor in win.seriko[id]:
                interval = actor.get_interval()
                if interval[:4] == "talk":
                    count = int(interval[5])
                    actor_id = actor.get_id()
                    break
            if count != None and self.text_count[self.script_side] >= count:
                self.surface.invoke(self.script_side, actor_id)
                self.text_count[self.script_side] = 0
            if self.script_speed > 0:
                self.script_wait = time.time() + self.script_speed * 0.02
            return
        node = self.processed_script.pop(0)
        if node[0] == ninix.script.SCRIPT_TAG:
            name, args = node[1], node[2:]
            if name == r"\e":
                win = self.surface.window[self.script_side]
                id = self.surface.get_surface(self.script_side)
                for actor in win.seriko[id]:
                    if actor.get_interval() == "yen-e":
                        actor_id = actor.get_id()
                        self.surface.invoke(self.script_side, actor_id)
                        break
                self.reset_script()
                self.__balloon_life = self.BALLOON_LIFE
            elif name in [r"\0", r"\h"]:
                ##self.balloon.show(0)
                self.script_side = 0
            elif name in [r"\1", r"\u"]:
                ##self.balloon.show(1)
                self.script_side = 1
            elif name == r"\4":
                if self.script_side == 0:
                    sw, sh = self.surface.get_surface_size(1)
                    sx, sy = self.surface.get_position(1)
                else:
                    sw, sh = self.surface.get_surface_size(0)
                    sx, sy = self.surface.get_position(0)
                w, h = self.surface.get_surface_size(self.script_side)
                x, y = self.surface.get_position(self.script_side)
                scrn_w = gtk.gdk.screen_width()
                scrn_h = gtk.gdk.screen_height() - self.bottom_margin
                if sx + sw / 2 > scrn_w / 2:
                    new_x = sx - w - int(scrn_w/20)
                else:
                    new_x = sx + sw + int(scrn_w/20)
                new_x = max(new_x, 0)
                new_x = min(new_x, scrn_w - w)
                if x > new_x:
                    step = -10
                else:
                    step = 10
                if abs(sx - new_x) < abs(sx - x):
                    return
                for current_x in range(x, new_x, step):
                    self.surface.set_position(self.script_side, current_x, y)
                    self.balloon.reset_balloon(0)
                    self.balloon.reset_balloon(1)
                self.surface.set_position(self.script_side, new_x, y)
                self.balloon.reset_balloon(0)
                self.balloon.reset_balloon(1)
            elif name == r"\5":
                if self.script_side == 0:
                    sw, sh = self.surface.get_surface_size(1)
                    sx, sy = self.surface.get_position(1)
                else:
                    sw, sh = self.surface.get_surface_size(0)
                    sx, sy = self.surface.get_position(0)
                w, h = self.surface.get_surface_size(self.script_side)
                x, y = self.surface.get_position(self.script_side)
                scrn_w = gtk.gdk.screen_width()
                scrn_h = gtk.gdk.screen_height() - self.bottom_margin
                if sx + sw/2 > x and sx + sw/2 < x + w or \
                   x + w/2 > sx and x + w/2 < sx + sw:
                    return
                if sx + sw / 2 > x + w / 2:
                    new_x = sx - w / 2 + 1
                else:
                    new_x = sx + sw - w / 2 - 1
                new_x = max(new_x, 0)
                new_x = min(new_x, scrn_w - w)
                if x > new_x:
                    step = -10
                else:
                    step = 10
                for current_x in range(x, new_x, step):
                    self.surface.set_position(self.script_side, current_x, y)
                    self.balloon.reset_balloon(0)
                    self.balloon.reset_balloon(1)
                self.surface.set_position(self.script_side, new_x, y)
                self.balloon.reset_balloon(0)
                self.balloon.reset_balloon(1)
            elif name == r"\s":
                id = args[0]
                if id == "-1":
                    self.surface.hide(self.script_side)
                else:
                    self.surface.set_surface(self.script_side, id)
                    self.surface.show(self.script_side)
                    self.surface.raise_(self.script_side)
                    self.balloon.raise_(self.script_side)
                    self.notify_event("OnSurfaceChange",
                                      self.surface.get_surface(0),
                                      self.surface.get_surface(1))
                    self.position_balloons()
            elif name == r"\b":
                if args[0] == "-1":
                    self.balloon.hide(self.script_side)
                else:
                    try:
                        id = int(args[0]) / 2
                    except ValueError:
                        id = 0
                    else:
                        self.balloon.set_balloon(self.script_side, id)
            elif name == r"\_b":
                filename, x, y = self.expand_meta(args[0]).split(',')
                filename = filename.lower()
                path = os.path.join(self.ghost.get_prefix(), 'ghost/master',
                                    filename.replace('\\', '/'))
                if os.path.isfile(path):
                    self.balloon.append_image(self.script_side, path, x, y)
            elif name == r"\n":
                if len(args) > 0 and self.expand_meta(args[0]) == 'half':
                    self.balloon.append_text(self.script_side, u"\n[half]")
                else:
                    self.balloon.append_text(self.script_side, u"\n")
            elif name == r"\c":
                self.balloon.clear_text(self.script_side)
            elif name in [r"\w", r"\_w"]:
                if not self.quick_session and self.script_speed >= 0:
                    if name == r"\w":
                        unit = 0.05  # 50ms
                    else:
                        unit = 0.001  # 1ms
                    try:
                        amount = int(args[0]) * unit - 0.01
                    except ValueError:
                        amount = 0
                    if amount > 0:
                        self.script_wait = time.time() + amount
            elif name == r"\t":
                self.time_critical_session = not self.time_critical_session
            elif name == r"\_q":
                self.quick_session = not self.quick_session
            elif name == r"\_s":
                self.set_synchronized_session(not self.synchronized_session)
            elif name == r"\_e":
                self.balloon.hide(self.script_side)
                self.balloon.clear_text(self.script_side)
            elif name == r"\q":
                newline_required = 0
                if len(args) == 3: # traditional syntax
                    num, id, text = args
                    newline_required = 1
                else: # new syntax
                    text, id = args
                text = self.expand_meta(text)
                self.balloon.append_link(self.script_side, id, text, newline_required)
                self.script_mode = self.SELECT_MODE
            elif name == r"\URL":
                text = self.expand_meta(args[0])
                if len(args) == 1:
                    link = text
                else:
                    link = "#cancel"
                self.balloon.append_link(self.script_side, link, text)
                for i in range(1, len(args), 2):
                    link = self.expand_meta(args[i])
                    text = self.expand_meta(args[i+1])
                    self.balloon.append_link(self.script_side, link, text)
                self.script_mode = self.SELECT_MODE
            elif name == r"\_a":
                if self.anchor:
                    id = self.anchor[0]
                    text = self.anchor[1]
                    self.balloon.append_link(self.script_side, id, text)
                    self.anchor = None
                else:
                    id = args[0]
                    self.anchor = [('anchor', id), '']
            elif name == r"\x":
                if self.script_mode == self.BROWSE_MODE:
                    self.script_mode = self.PAUSE_MODE
                    self.balloon.window[self.script_side].redraw()
                    self.balloon.append_text(self.script_side, u"\n")
            elif name == r"\a": # obsolete # only for old SHIORI
                self.start_script(self.ghost.getaistringrandom())
            elif name == r"\i":
                try:
                    id = int(args[0])
                except ValueError:
                    pass
                else:
                    self.surface.invoke(self.script_side, id)
            elif name == r"\j":
                id = args[0]
                if self.is_URL(id):
                    self.launch_browser(id)
            elif name == r"\-":
                self.ghost.quit()
            elif name == r"\+":
                self.ghost.select_ghost(1)
            elif name == r"\_+":
                self.ghost.select_ghost(0)
            elif name == r"\m":
                self.write_sstp_handle(self.expand_meta(args[0]))
            elif name == r"\&":
                if self.name2codepoint is not None:
                    text = unichr(self.name2codepoint.get(args[0]))
                else:
                    text = None
                if text is None:
                    text = "?"
                self.balloon.append_text(self.script_side, text)
            elif name == r"\_m":
                try:
                    num = int(args[0], 16)
                except ValueError:
                    num = 0
                if num >= 0x20 and num <= 0x7e:
                    text = chr(num)
                else:
                    text = "?"
                self.balloon.append_text(self.script_side, text)
            elif name == r"\_u":
                self.balloon.append_text(self.script_side, u"?")
            elif name == r"\_v":
                filename = self.expand_meta(args[0])
                filename = filename.lower()
                path = os.path.join(self.ghost.get_prefix(), 'ghost/master', filename)
                if os.path.isfile(path):
                    for regex, command in self.helpers:
                        if regex.search(path):
                            self.execute_command(command, path)
                            break
            elif name == r"\!" and len(args) > 0:
                argc = len(args)
                args = [self.expand_meta(s) for s in args]
                if args[0] == "raise" and argc >= 2:
                    apply(self.notify_event, tuple(args[1:10]))
                elif args[0:2] == ["open", "browser"] and argc > 2:
                    self.launch_browser(args[2])
                elif args[0:2] == ["open", "communicatebox"]:
                    if not self.passivemode:
                        self.open_communicatebox()
                elif args[0:2] == ["open", "teachbox"]:
                    if not self.passivemode:
                        self.open_teachbox()
                elif args[0:2] == ["open", "inputbox"] and argc > 2:
                    if not self.passivemode:
                        if argc > 4:
                            self.open_inputbox(args[2], args[3], args[4])
                        elif argc == 4:
                            self.open_inputbox(args[2], args[3])
                        else:
                            self.open_inputbox(args[2])
                elif args[0:2] == ["open", "configurationdialog"]:
                    self.ghost.edit_preferences()
                elif args[0:2] == ["change", "ghost"] and argc > 2:
                    if args[2] == "random":
                        self.ghost.select_ghost(0, 0)
                    else:
                        self.ghost.select_ghost_by_name(args[2], 0)
                elif args[0:1] == ["updatebymyself"]:
                    if not self.busy():
                        self.ghost.update()
                elif args[0:1] == ["vanishbymyself"]:
                    self.vanished = 1
                    count = self.get_ghost_vanished_count()
                    self.set_ghost_vanished_count(count + 1)
                    self.set_ghost_time(0)
                    self.ghost.app.stop_sakura(self.ghost.app.vanish_sakura)
                elif args[1:2] == ["repaint"]:
                    if args[0:1] == ["lock"]:
                        self.lock_repaint = 1
                    elif args[0:1] == ["unlock"]:
                        self.lock_repaint = 0
                elif args[1:2] == ["passivemode"]:
                    if args[0:1] == ["enter"]:
                        self.passivemode = 1
                    elif args[0:1] == ["leave"]:
                        self.passivemode = 0
                elif args[0:2] == ["set", "alignmentondesktop"]:
                    if args[2] == "bottom":
                        if self.synchronized_session:
                            self.align_bottom(0)
                            self.align_bottom(1)
                        else:
                            self.align_bottom(self.script_side)
                    elif args[2] == "top":
                        if self.synchronized_session:
                            self.align_top(0)
                            self.align_top(1)
                        else:
                            self.align_top(self.script_side)
                    self.position_balloons()
                elif args[0:2] == ["set", "alignmenttodesktop"]:
                    if args[2] == "free":
                        if self.synchronized_session:
                            self.surface.set_alignment(0, 2)
                            self.surface.set_alignment(1, 2)
                        else:
                            self.surface.set_alignment(self.script_side, 2)
                    self.position_balloons()
                elif args[0] == '*':
                    self.balloon.append_sstp_marker(self.script_side)
            elif name == r"\__c":
                self.open_communicatebox()
            elif name == r"\__t":
                self.open_teachbox()
        elif node[0] == ninix.script.SCRIPT_TEXT:
            text = self.expand_meta(node[1])
            if self.anchor:
                self.anchor[1] = self.anchor[1] + text
            elif not self.quick_session and self.script_speed >= 0:
                self.processed_text = text
            else:
                self.balloon.append_text(self.script_side, text)
    def open_communicatebox(self):
        if self.user_interaction:
            return
        self.balloon.show_communicatebox()
        self.user_interaction = 1
    def open_teachbox(self):
        if self.user_interaction:
            return
        self.notify_event("OnTeachStart")
        self.balloon.show_teachbox()
        self.user_interaction = 1
    def open_inputbox(self, symbol, limittime=-1, default=None):
        if self.user_interaction:
            return
        self.balloon.show_inputbox(symbol, limittime, default)
        self.user_interaction = 1
    def reset_script(self, reset_all=0):
        if reset_all:
            self.script_mode = self.BROWSE_MODE
            self.script_post_proc = None
        self.processed_script = None
        self.processed_text = ""
        self.time_critical_session = 0
        self.quick_session = 0
        self.set_synchronized_session(0)
        self.reset_idle_time()
    def set_synchronized_session(self, sync):
        if sync:
            self.synchronized_session = 1
        else:
            self.synchronized_session = 0
        self.balloon.synchronize(self.synchronized_session)
    def expand_meta(self, text_node):
        buffer = []
        for chunk in text_node:
            if chunk[0] == ninix.script.TEXT_STRING:
                buffer.append(chunk[1])
            elif chunk[1] == "%month":
                buffer.append(str(self.current_time[1]))
            elif chunk[1] == "%day":
                buffer.append(str(self.current_time[2]))
            elif chunk[1] == "%hour":
                buffer.append(str(self.current_time[3]))
            elif chunk[1] == "%minute":
                buffer.append(str(self.current_time[4]))
            elif chunk[1] == "%second":
                buffer.append(str(self.current_time[5]))
            elif chunk[1] in ["%username", "%c"]:
                buffer.append(self.get_username())
            elif chunk[1] == "%selfname":
                buffer.append(self.get_selfname())
            elif chunk[1] == "%selfname2":
                buffer.append(self.get_selfname2())
            elif chunk[1] == "%keroname":
                buffer.append(self.get_keroname())
            elif chunk[1] == "%friendname":
                buffer.append(self.get_friendname())
            elif chunk[1] == "%screenwidth":
                buffer.append(str(gtk.gdk.screen_width()))
            elif chunk[1] == "%screenheight":
                buffer.append(str(gtk.gdk.screen_height()))
            elif chunk[1] == "%et":
                buffer.append(unicode("%d万年" % self.current_time[7], 'UTF-8'))
            elif chunk[1] == "%exh":
                buffer.append(str(self.get_uptime()))
            elif chunk[1] in ["%ms", "%mz", "%ml", "%mc", "%mh", \
                              "%mt", "%me", "%mp", "%m?"]:
                buffer.append(self.ghost.getword(chunk[1][1:]))
            elif chunk[1] == "%dms":
                buffer.append(self.ghost.getdms())
            else: # %c, %songname
                buffer.append(chunk[1])
        return string.join(buffer, '')
    ###   SEND SSTP/1.3   ###
    def open_sstp_handle(self, path):
        handle = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            handle.connect(path)
        except socket.error:
            handle = None # discard socket object
            sys.stderr.write("cannot open Unix socket: %s\n" % path)
        return handle
    def _send_sstp_handle(self, data):
        r, w, e = select.select([], [self.sstp_handle], [], 0)
        if not w:
            return
        try:
            self.sstp_handle.send(data + "\n")
        except socket.error:
            pass
    def write_sstp_handle(self, data):
        if self.sstp_handle is None:
            return
        self._send_sstp_handle("+" + data)
        ##print "write_sstp_handle(%s)" % repr(data)
    def close_sstp_handle(self):
        if self.sstp_handle is None:
            return
        self._send_sstp_handle("-")
        ##print "close_sstp_handle()"
        try:
            self.sstp_handle.close()
        except socket.error:
            pass
        self.sstp_handle = None

class Ghost:
    def __init__(self, app, default_path, debug=0):
        self.app = app
        self.__sakura  = ninix.sakura.Sakura(debug)
        self.__sakura.set_ghost(self)
        saori_lib = ninix.dll.Library('saori', default_path, sakura=self.__sakura)
        self.__dll = ninix.dll.Library('shiori', default_path, saori_lib=saori_lib)
        self.debug = debug
        self.surface = ninix.surface.Surface(self.__sakura, debug)
        self.__sakura.set_surface(self.surface)
        self.__balloon = ninix.balloon.Balloon(self.__sakura, debug)
        self.__sakura.set_balloon(self.__balloon)
    def get_sakura(self): ## FIXME: this method should be deleted
        return self.__sakura
    def get_default_surface_scale(self):
        return self.__sakura.get_default_surface_scale()
    def get_default_script_speed(self):
        return self.__sakura.get_default_script_speed()
    def set_top_margin(self, margin):
        self.__sakura.set_top_margin(margin)
    def set_bottom_margin(self, margin):
        self.__sakura.set_bottom_margin(margin)
    def set_browser(self, command):
        self.__sakura.set_browser(command)
    def set_helpers(self, list):
        self.__sakura.set_helpers(list)
    def set_scale_balloon(self, flag):
        self.__sakura.set_scale_balloon(flag)
    def set_surface_scale(self, scale):
        self.__sakura.set_surface_scale(scale)
    def set_script_speed(self, speed):
        self.__sakura.set_script_speed(speed)
    def set_event_kill_list(self, list):
        self.__sakura.set_event_kill_list(list)
    def stop(self):
        self.__sakura.stop()
    def restart(self):
        self.__sakura.restart()
    def set_balloon(self, desc, balloon):
        self.__balloon.new(desc, balloon)
    def busy(self):
        return self.__sakura.busy()
    def set_balloon_fonts(self, fonts):
        self.__balloon.set_balloon_fonts(fonts)
    def get_balloon_fonts(self):
        return self.__balloon.get_balloon_fonts()
    def set_mouse_button1(self, button1):
        self.__sakura.set_mouse_button1(button1)
    def set_mouse_button3(self, button3):
        self.__sakura.set_mouse_button3(button3)
    def finalize(self):
        self.__sakura.finalize()
    def get_vanished_count(self): ## FIXME
        return self.__sakura.get_ghost_vanished_count()
    def set_vanished_count(self, count): ## FIXME
        self.__sakura.set_ghost_vanished_count(count)
    def set_ghost_time(self, time): ## FIXME
        self.__sakura.set_ghost_time(time)
    def save_history(self): ## FIXME
        self.__sakura.save_history()
    def unload_shiori(self):
        self.shiori.unload()
    def get_selfname(self):
        return self.__sakura.get_selfname()
    def get_sstp_port(self):
        return self.__sakura.get_sstp_port()
    def add_sstp_server(self, server):
        self.__sakura.add_sstp_server(server)
    def add_directsstp_server(self, server, path):
        self.__sakura.add_directsstp_server(server, path)
    def notify_shell_changing(self, name, proc):
        self.__sakura.notify_shell_changing(name, proc)
    def notify_ghost_changing(self, name, method, proc):
        self.__sakura.notify_ghost_changing(name, method, proc)
    def notify_close(self, proc):
        self.__sakura.notify_close(proc)
    def notify_shell_changed(self, name):
        self.__sakura.notify_shell_changed(name)
    def notify_ghost_changed(self, name, vanished=0):
        self.__sakura.notify_ghost_changed(name, vanished)
    def notify_ninix_reloading(self, proc):
        self.__sakura.notify_ninix_reloading(proc)
    def notify_ninix_reloaded(self):
        self.__sakura.notify_ninix_reloaded()
    def notify_vanish_selecting(self):
        self.__sakura.notify_vanish_selecting()
    def notify_vanish_cancel(self):
        self.__sakura.notify_vanish_cancel()
    def notify_vanish_selected(self, proc):
        self.__sakura.notify_vanish_selected(proc)
    def get_own_balloon_name(self):
        if self.own_balloon:
            desc, balloon = self.own_balloon
            name = desc.get('name', unicode(_("Own Balloon"), 'utf-8'))
        else:
            name = None
        return name
    def get_current_balloon_name(self):
        return self.__balloon.get_balloon_name()
    def new(self, desc, shiori_dir, use_makoto, balloon, prefix, shiori_dll, shiori_name):
        self.shiori = None
        self.desc = desc
        self.own_balloon = balloon
        name = (shiori_dll, shiori_name)
        self.shiori = self.__dll.request(name)
        if self.shiori and self.shiori.load(shiori_dir):
            if getattr(self.shiori, 'show_description', None):
                self.shiori.show_description()
        else:
            sys.stderr.write("cannot load SHIORI(%s) (abort)\n" % shiori_name)
            sys.exit(1)
        self.use_makoto = use_makoto
        self.prefix = prefix
        self.updateman = ninix.update.NetworkUpdate(self.__sakura)
        icon = self.desc.get("icon", None)
        if icon != None:
            icon_path = os.path.join(shiori_dir, icon)
            if not os.path.exists(icon_path):
                icon_path = None
        else:
            icon_path = None
        self.__sakura.surface.set_icon(icon_path)        
    def start(self):
        self.__sakura.start()
        self.set_timeout()
    def get_prefix(self):
        return self.prefix
    def translate(self, s):
        if s is not None:
            if self.use_makoto:
                s = ninix.makoto.execute(s)
            else:
                r = self.get_event_response("OnTranslate", s, translate=0)
                if r:
                    s = r
        return s
    def getaistringrandom(self): # obsolete
        result = self.get_event_response("OnAITalk")
        return self.translate(result)
    def getdms(self):
        result = self.get_event_response('dms')
        return self.translate(result)
    def getword(self, type):
        result = self.get_event_response('%s' % type)
        return self.translate(result)
    def getstring(self, name):
        return self.get_event_response(name)
    def get_name(self):
        return self.desc.get("name", unicode(_("Sakura&Unyuu"), 'utf-8'))
    def get_username(self):
        return self.desc.get("user.defaultname", unicode(_("User"), 'utf-8'))
    def get_selfname(self):
        return self.desc.get("sakura.name", unicode(_("Sakura"), 'utf-8'))
    def get_selfname2(self):
        return self.desc.get("sakura.name2", unicode(_("Sakura"), 'utf-8'))
    def get_keroname(self):
        return self.desc.get("kero.name", unicode(_("Unyuu"), 'utf-8'))
    def get_friendname(self):
        return self.desc.get("sakura.friend.name", unicode(_("Tomoyo"), 'utf-8'))
    def get_event_response(self, event,
                           ref0=None, ref1=None, ref2=None, ref3=None,
                           ref4=None, ref5=None, ref6=None, ref7=None,
                           ref_ext=[], type='GET', translate=1):
        ref = [ref0, ref1, ref2, ref3, ref4, ref5, ref6, ref7]
        ref.extend(ref_ext)
        header = '%s SHIORI/3.0\r\n' % type + \
                 'ID: %s\r\n' % event + \
                 'Sender: ninix-aya\r\n' \
                 'SecurityLevel: local\r\n'
        for i in range(len(ref)):
            if ref[i] != None:
                header += 'Reference' + str(i) + ': ' + str(ref[i]) + '\r\n'
        header += '\r\n'
        header = header.encode(self.__sakura.charset, 'ignore')
        response = self.shiori.request(header)
        if self.__sakura.cantalk:
            result, to = self.get_value(response)
            if translate:
                result = self.translate(result)
        else:
            result, to = '', None
        if result == None:
            result = ''
        if to and result:
            header = 'COMMUNICATE SSTP/1.1\r\n' + \
                     'Sender: %s\r\n' % self.__sakura.get_selfname() + \
                     'Sentence: %s\r\n' % result + \
                     'Charset: %s\r\n\r\n' % self.__sakura.charset
            header = header.encode(self.__sakura.charset, 'ignore')
            self.__sakura.communicate.send_message(to, header)
        return result
    def get_value(self, response):
        header = StringIO.StringIO(response)
        result = {}
        to = None
        for line in header:
            if line[-1] == '\n':
                line = line[:-1]
            line = line.strip()
            if not line:
                continue
            colon = line.find(':')
            if colon >= 0:
                key = line[:colon].strip()
                result[key] = line[colon+1:].strip()
            else:
                continue
        for key in result.keys():
            result[key] = unicode(result[key], self.__sakura.charset, 'ignore')
        if result.has_key('Reference0'):
            to = result['Reference0']
        if result.has_key('Value'):
            return result['Value'], to
        else:
            return None, to
    def teach(self, word):
        result = self.get_event_response('OnTeach', word)
        return self.translate(result)
    def update(self):
        if self.updateman.is_active():
            return
        homeurl = self.getstring("homeurl")
        if not homeurl:
            self.__sakura.start_script(r"\t\h\s[0]" + unicode(_("I'm afraid I don't have Network Update yet."), 'utf-8') + r"\e")
            self.__sakura.balloon.hide_sstp_message()
            return
        ghostdir = self.get_prefix()
        print "homeurl =", homeurl
        print "ghostdir =", ghostdir
        self.updateman.start(homeurl, ghostdir)
    def quit(self):
        self.app.quit()
    def edit_preferences(self):
        self.app.edit_preferences(None)
    def reload(self):
        self.app.reload()
    def select_ghost(self, sequential, event=1):
        self.app.select_ghost(sequential, event)
    def select_ghost_by_name(self, name, event=1):
        self.app.select_ghost_by_name(name, event)
    def select_balloon(self, widget, item):
        if not widget.active:
            return
        ## FIXME
        ##if self.__sakura.busy():
        ##    menuitem, hid = self.balloon_menu[self.current_balloon]
        ##    menuitem.activate()
        ##    gtk.gdk.beep()
        ##    return
        if self.own_balloon:
            desc, balloon = self.own_balloon
            name = self.get_own_balloon_name()
        else:
            name = None
        if item != name:
            desc, balloon = self.app.get_balloon_description(item)
        self.__balloon.hide(0)
        self.__balloon.hide(1)
        self.__balloon.new(desc, balloon)
        self.__sakura.set_balloon(self.__balloon)
        self.__balloon.set_balloon(0, 0)
        self.__balloon.set_balloon(1, 0)
        print "balloon", item
    ###   TIMEOUT   ###
    def set_timeout(self):
        gtk.timeout_add(10, self.do_idle_tasks) # 10ms
    reload_event = None
    def do_idle_tasks(self):
        if self.reload_event and not self.__sakura.busy() and \
           not (self.__sakura.processed_script or self.__sakura.processed_text):
            self.__sakura.hide_all()
            sys.stdout.write('reloading....\n')
            self.shiori.unload()
            self.app.reload_current_sakura()
            self.__sakura.restart()
            sys.stdout.write('done.\n')
            apply(self.__sakura.enqueue_event, self.reload_event)
            self.reload_event = None
        self.__sakura.update_surface()
        self.__sakura.receive_sstp_request()
        # continue network update (enqueue events)
        if self.updateman.is_active():
            self.updateman.run()
            while 1:
                event = self.updateman.get_event()
                if not event:
                    break
                if event[0] == 'OnUpdateComplete' and event[1] == 'changed':
                    self.reload_event = event
                else:
                    apply(self.__sakura.enqueue_event, event)
        self.__sakura.process_script()
        self.set_timeout()
    def launch_browser(self, event, args):
        title, url = args
        if self.__sakura.is_URL(url):
            self.__sakura.launch_browser(url)
        self.__sakura.enqueue_event("OnRecommandedSiteChoice", title, url)
    def close(self, action, widget):
        self.app.close(None)
    def about(self, action, widget):
        if self.__sakura.busy():
            gtk.gdk.beep()
            return
        self.__sakura.notify_about()
    def network_update(self, action, widget):
        if self.__sakura.busy():
            gtk.gdk.beep()
            return
        self.update()
    def edit_preferences(self, action, widget):
        self.app.edit_preferences(None)
    def reload(self, action, widget):
        self.app.reload()
    def show_usage(self, action, widget):
        self.app.show_usage(widget)
    def vanish(self, action, widget):
        self.app.vanish(None)

def test():
    pass

if __name__ == "__main__":
    test()
